import copy
import math

class CacheLine(object):
    def __init__(self):
        self.tag = 0
        self.setnum = 0
        self.valid = False

class Cache(object):
    def __init__(self, bsize, assoc, size, vbsize=0):
        self.bsize = bsize
        self.assoc = assoc
        self.size = size
        self.num_sets = (size / self.assoc) / self.bsize
        self.vbsize = vbsize
        assert self.num_sets * self.assoc * self.bsize == size, \
            "Cache size is not evenly divisible by block size " \
            "and/or associativity!"
                
        # bits used for block offset
        self.block_offset_bits = int( math.log( self.bsize, 2 ) )
        # bits used for index
        self.index_bits = int( math.log( self.num_sets, 2 ) )

        self.reset()

    def has_victim_buffer(self):
        return self.victim_buffer is not None

    def setnum(self, address):
        return (address >> self.block_offset_bits) % self.num_sets

    def tag(self, address):
        return address >> (self.block_offset_bits + self.index_bits)

    def access(self, address):
        """Requests the given address from the cache. Returns True for 
        a hit and False for a miss.
        """
        setnum = self.setnum( address )
        tag = self.tag( address )
        set = self.cache[ setnum ]
        # search for our cache line
        hits = [ cl for cl in set if cl.valid and cl.tag == tag ]
        if len( hits ) > 0:
            assert 1 == len( hits ), "Duplicate cache line!"
            hit = hits[0]
            # mark line as MRU
            set.remove( hit )
            set.append( hit )
            return True

        # we missed in the cache - find LRU line to evict
        lru = set[0]
        
        # check victim buffer
        if self.victim_buffer is not None:
            hits = [ cl for cl in self.victim_buffer 
                     if cl.valid and cl.tag == tag and cl.setnum == setnum]
            if len( hits ) > 0:
                assert 1 == len( hits ), "Duplicate cache lines!"

                # remove hit from victim buffer
                self.victim_buffer.remove( hits[0] )

                # evict LRU line to head of the victim buffer
                self.victim_buffer.append( copy.deepcopy(lru) )

                # move victim buffer hit back into cache by overwriting lru
                lru.tag = tag
                lru.setnum = setnum
                lru.valid = True
                # mark as MRU
                set.remove( lru )
                set.append( lru )
                return True


        # a true miss :-(
        
        # evict LRU line to head of the victim buffer
        if lru.valid and self.victim_buffer is not None:
            self.victim_buffer.pop( 0 )
            self.victim_buffer.append( copy.deepcopy(lru) )
            pass

        # fill with new line by overwriting lru
        lru.tag = tag
        lru.setnum = setnum
        lru.valid = True
        # mark as MRU
        set.remove( lru )
        set.append( lru )
        return False

    def reset(self):
        """Empty the cache and the victim buffer (if any)."""
        # setup victim buffer, if we have one
        self.victim_buffer = None
        if self.vbsize > 0: 
            self.victim_buffer = []
            for _ in range(self.vbsize):
                self.victim_buffer.append( CacheLine() )
        
        # setup cache itself
        self.cache = []
        for _ in range(self.num_sets):
            set = []
            for __ in range(self.assoc):
                set.append( CacheLine() )
                pass
            self.cache.append( set )
        
